/*
 * Gnop - a PS2Linux based Pong clone, using SPS2
 * (or, exactly how much fun can you have with 13 coloured boxes?)
 * Gnop could also stand for "GNOP's not 'only' Pong"
 *
 * Code has been adapted as simply as possible by Eratosthenes from
 * Dr. H. S. Fortuna's tutorial code "A simple game" available from 
 * http://www.hsfortuna.pwp.blueyonder.co.uk
 *
 * Eratosthenes wishes to thank Dr. Fortuna for his valuable contribution
 * to the PS2Linux community by making his tutorials open to all when he
 * could have kept them privately at his university.  Also, Henry helped
 * greatly by providing valuable feedback on the folowing code before its
 * release
 *
 * Eratosthenes also wishes to thank Sauce et al for the SPS2 module in
 * the first place, which makes the code below so easy to write
 *
 * Finally, Eratosthenes wishes he chose a name that was easier to spell.
 *
 * TECHNICAL NOTES
 *
 * Screen orientation:
 * It may confuse some people to find that the original tutorial had a 
 * small bug in the description of the y co-ordinate of each box.  (x, y)
 * was said to be the top-left co-ordinate of each box, where in fact it 
 * is the bottom left, as the PS2s screen is inverted (in the Cartesian 
 * sense) so the as x increases, a point moves from left to right, but as 
 * y increases, a point moves from top to bottom.
 *
 * Collision detection:
 * There is a small problem with the collision-detection code.
 * Basically, it doesn't account for the possibility that 2 boxes might be
 * moving so fast that they pass through each other completely in a single 
 * frame.  I don't think that'll happen in a sane version of Pong (things 
 * would have to move really very fast so I've left it in.
 *
 * IMPROVING THE GAME
 * 
 * To put it simply, there are an infinite number of possible improvements 
 * to this game.  Here are just a few you could have some fun with.
 *
 * * Variable ball movment/speed
 * If players are too good at the initial speed settings, the ball needs 
 * to be sped up as the game increases.  More worryingly, the ball comes 
 * off the bat at a totally predictable angle every time, which doesn't 
 * make for a very good game.
 *
 * * Variable controls
 * It's always nice to allow a player to use their own choice of buttons 
 * to control the game, but in this case it should be possible to allow 2 
 * players to play from a single controller, as some people do not have a 
 * second controller.  Currently the game exits with a harsh error message
 * if 2 controllers are not found.
 *
 * * Analogue controls (pressure + sticks)
 * Part of the above section really - analogue controls are sometimes 
 * better than digital controls, and the player should be given a choice.
 * For instance, as currently set up, the player must hold X to move 
 * faster, why introduce this extra button when you could merely have the
 * player push the D-pad harder, which some may argue is more natural?
 *
 * * Title screen
 * Again, it's often polite to wait until players are ready before 
 * starting a game.  Whilst the one-second delay helps, a title screen 
 * would help greatly here.  I strongly doubt anyone could possibly have
 * grown up without hearing of pong, or without instantly realising how 
 * to play the game, it would also be polite to have some kind of how to
 * play demo feature in the intro.
 *
 * * End game
 * To comlpement the title screen, the game shouldn't just end when one 
 * player has lost 3 lives.  We should first inform the players who has 
 * won, and then we should allow the choice of whether to play another 
 * game or not before exiting.
 *
 * * 1-player mode
 * How about a computer player to play against?  This would be quite 
 * difficult, as we need to create an AI that plays well, but has enough 
 * imperfections to be considered beatable.  We could have a series of 
 * opponents to beat, each with different names, or we could simply pick a 
 * difficulty level.
 *
 * * Different play styles
 * I've put together a very simple version of pong here, but there are 
 * several well-known variants out there, such as soccer (where each 
 * player has 2 bats, effectively one striker and one goalkeeper).  How 
 * about giving each player 2 bats, one which moves vertically, the other 
 * horizontally, and give one player the task of defending both the top 
 * and left walls?  How about having targets behind the bat that need to 
 * be knocked down?  Or have a goal that shrinks when your losing, to give 
 * the losing player a crucial advantage?
 *
 * * Story mode?
 * Think of your favourite big budget game, and most of them have some 
 * kind of story mode.  How about introducing some pong based challenges, 
 * and awarding the player medals for completing them, a certain number of 
 * which unlock new challenges and playing styles?
 *
 * * On-line play?
 * OK, I know it's 'only' Pong, but why not test out your network coding 
 * skills by allowing the 2 players to be on different PS2's?  This will 
 * obviously require 2 PS2s, or a willing friend, to be tested though...
 *
 * * More that I've missed?
 * The point I'm trying to make is that the number of improvements you 
 * could make is absolutely huge, and this is 'only' Pong.  Let your 
 * imagination run wild!
 *
 * And now - the code... best viewed with a tab width of 8
 */

#include <stdio.h>		//So we can printf errors to the console
#include <assert.h>		//For 'invisible' debugging
#include <sps2lib.h>
#include <sps2tags.h>
#include <sps2util.h>
#include <sps2regstructs.h>
#include "pad.h"

// A structure type to hold box info
typedef struct
{
	int x, y;				//Bottom LH corner
	int z;					//z depth
	int w, h;				//width and height;
	unsigned char r, g, b;			//colour
	int xvel;				//Horizontal velocity
	int yvel;				//Vertical velocity
}Ps2Box_t;

//I've used an enum to keep track of which of the 13 boxes are which.  I 
//could just have used Box[0], Box[1] and so on - but I'd probably 
//forget which one was the ball :)
//The last term in the enum is NUMBOX - this gets the compiler to count 
//the number of boxes for me
enum {UPPER_WALL, LOWER_WALL, LEFT_BAT, RIGHT_BAT, LEFT_LIFE_1, LEFT_LIFE_2, LEFT_LIFE_3, RIGHT_LIFE_1, RIGHT_LIFE_2, RIGHT_LIFE_3, BOX_OUTER, BOX_INNER, BALL, NUMBOX};

// Declare the boxes (Global for simplicity just now!)
Ps2Box_t Box[NUMBOX];
int SPS2Descriptor;     // SPS2 handle
sps2Memory_t * pMemory;	// Info struct about allocated memory
int counter;            // This counter value is updated each frame

// Function prototypes
void GameLogic(void);
void Render(void);
int BuildGSPacket(void * pMem);
void ConfigureBox(Ps2Box_t * B, int x, int y, int z, int w, int h, 
			u_char r, u_char g, u_char b, int xvel, int yvel);

int main(void)
{
  	// Set the default parameters for the boxes to be drawn
        // Fiddle with some of these values and see what happens!
        // Look at the ConfigureBox prototype above to determine what 
        // order the parameters are in
  	ConfigureBox(&Box[UPPER_WALL], -220, -200, 2000, 440, 20, 128, 0, 0, 0, 0);
	ConfigureBox(&Box[LEFT_BAT], -220, 50, 2000, 20, 100, 0, 128, 0, 0, 0);
	ConfigureBox(&Box[LOWER_WALL], -220, 220, 2000, 440, 20, 0, 0, 128, 0, 0);
	ConfigureBox(&Box[RIGHT_BAT], 200, 50, 2000, 20, 100, 0, 128, 128, 0, 0);
	ConfigureBox(&Box[BOX_OUTER], -240, 240, 1000, 480, 480, 128, 128, 128, 0, 0);
	ConfigureBox(&Box[BOX_INNER], -230, 230, 1500, 460, 460, 64, 64, 64, 0, 0);
	ConfigureBox(&Box[BALL], -2, 2, 2000, 4, 4, 0, 0, 0, 2, 1);
	ConfigureBox(&Box[LEFT_LIFE_1], -255, 20, 2000, 10, 10, 0, 128, 0, 0, 0);
	ConfigureBox(&Box[LEFT_LIFE_2], -255, 5, 2000, 10, 10, 0, 128, 0, 0, 0);
	ConfigureBox(&Box[LEFT_LIFE_3], -255, -10, 2000, 10, 10, 0, 128, 0, 0, 0);
	ConfigureBox(&Box[RIGHT_LIFE_1], 245, 20, 2000, 10, 10, 0, 128, 128, 0, 0);
	ConfigureBox(&Box[RIGHT_LIFE_2], 245, 5, 2000, 10, 10, 0, 128, 128, 0, 0);
	ConfigureBox(&Box[RIGHT_LIFE_3], 245, -10, 2000, 10, 10, 0, 128, 128, 0, 0);		

	// Initialise sps2
  	SPS2Descriptor = sps2Init();       	

	// Initialise the Pads
	if(pad_init (PAD_0 | PAD_1, PAD_INIT_ANALOGUE ) == 0)
	{
		printf("Unable to initialise both pads, and this is a 2-player game!\n");
		pad_cleanup (PAD_0 | PAD_1);
		exit(0);
	}

  	// Allocate 4k of uncached SPS2 memory
  	pMemory=sps2Allocate(4096, SPS2_MAP_BLOCK_4K | SPS2_MAP_UNCACHED, SPS2Descriptor);

  	// Initialise the screen.  
  	sps2UScreenInit(0);

	// Render loop
	// Press Start and select on Pad 1 to end the program
  	do 
  	{
  		// Update the pads
  		pad_update (PAD_0 | PAD_1);
  		
  		// Perform some game logic
  		GameLogic();
  		
    	// Clear the screen to blue.
    	sps2UScreenClear(0x00, 0x00, 0x40);
		
		// Send data to GS
		Render();
    	
    	// Swap display buffers
    	sps2UScreenSwap();
    	
  	}while(!((pad[0].buttons & PAD_START)&&(pad[0].buttons & PAD_SELECT)));

	// Shut down the Pads  
  	pad_cleanup (PAD_0 | PAD_1);
  
  	// Shut down the screen
  	sps2UScreenShutdown();

  	// Free the memory
  	sps2Free(pMemory);

  	// and close sps2
  	sps2Release(SPS2Descriptor);
  	return 0;
}

// Build the GS packet. Function takes a pointer to the allocated memory
// and returns the qwords written to this memory
int BuildGSPacket(void * pMem) 
{
	int iQWC;					//qword count
	// GIFTag pointer
  	sps2GIFTag_t * pGIFTag = (sps2GIFTag_t *) pMem;
  	// GIF Register pointer
  	sps2GIFPackedRegister_t * pReg; 
  	
  	// Prepare the GIF tag  
  	pGIFTag->i128 = 0;     			// Clear it
  	pGIFTag->s.NLOOP = 0;  			// Set to 0 and increment as required
  	pGIFTag->s.PRE =1 ;      		// provide a valid PRIM value
  	pGIFTag->s.EOP = 1;      		// End of packet
  	pGIFTag->s.PRIM = 0xE;    		// 6 = Sprite | 8 = gourard shading
  	pGIFTag->s.FLG = GIF_FLG_PACKED;// Packed format 
  	pGIFTag->s.NREG = 0;     		// Start at zero and inc as required
  
  	// Add RGBAQ register.
  	SPS2_SET_GIF_REG(*pGIFTag, pGIFTag->s.NREG++, GIF_REG_RGBAQ);
  	// Add XYZ2 register.
  	SPS2_SET_GIF_REG(*pGIFTag, pGIFTag->s.NREG++, GIF_REG_XYZ2);
	
	// Set pReg to qword after the GIFTag
  	pReg = (sps2GIFPackedRegister_t *) (pGIFTag + 1 );
  
  	// Loop through all vertices
  	for (int i=0;i<NUMBOX;i++)
	{
    	pGIFTag->s.NLOOP++;  			// Increment NLOOP in GIFTag
    
    	pReg->i128=0;   			// Blank out this value
    	// Set RGBA values
    	pReg->RGBAQ.s.R = Box[i].r;
    	pReg->RGBAQ.s.G = Box[i].g;
    	pReg->RGBAQ.s.B = Box[i].b;
    	pReg->RGBAQ.s.A = 0x80; 		// Alpha = 128
    	
    	pReg++;					// Increment register data pointer
    
    	pReg->i128=0;  				// Blank out this value
    	// Set XYZ values
    	pReg->XYZ2.s.X = (2048 + Box[i].x)<< 4;
    	pReg->XYZ2.s.Y = (2048 + Box[i].y)<< 4;
    	pReg->XYZ2.s.Z = Box[i].z;
    
    	pReg++;					// Increment register data pointer
    	
    	pGIFTag->s.NLOOP++;  			// Increment NLOOP in GIFTag
    
    	pReg->i128=0;   			// Blank out this value
    	// Set RGBA values
    	pReg->RGBAQ.s.R = Box[i].r;
    	pReg->RGBAQ.s.G = Box[i].g;
    	pReg->RGBAQ.s.B = Box[i].b;
    	pReg->RGBAQ.s.A = 0x80; 		// Alpha = 128
    	
    	pReg++;					// Increment register data pointer
    
    	pReg->i128=0;  				// Blank out this value
    	// Set XYZ values
    	pReg->XYZ2.s.X = (2048 + Box[i].x + Box[i].w)<< 4;
    	pReg->XYZ2.s.Y = (2048 + Box[i].y - Box[i].h)<< 4;
    	pReg->XYZ2.s.Z = Box[i].z;
    
    	pReg++;					// Increment register data pointer
    }

  	// Determine how many bytes have been written.
  	// Shift right by 4 to give qwords written
  	iQWC = ((int) pReg - (int) pMem) >> 4;
  	
  	// assert that qwords written is less than allocated memory
  	assert(iQWC < (4096 >> 4));
  	
  	//return qword count
  	return iQWC;
}

// Configure the box parameters
void ConfigureBox(Ps2Box_t * B, int x, int y, int z, int w, int h, 
			u_char r, u_char g, u_char b, int xvel, int yvel)
{
	B->x = x;
	B->y = y;
	B->z = z;
	B->w = w;
	B->h = h;
	B->r = r;
	B->g = g;
	B->b = b;
	B->xvel = xvel;
	B->yvel = yvel;
}


// Send the packet to the GS
void Render(void)
{
	int iQWC;
	// Channel control register used to control the DMA transfer
  	Dn_CHCR_t chcr;	
  	
  	// Set the channel control register values.
  	// These will not change.
  	chcr.i32 = 0;                  // Clear everything
  	chcr.s.MOD = CHCR_MOD_NORMAL;  // Normal mode transfer
  	chcr.s.STR = 1;                // STR=1 to start transfer
	
	// Build the GSPacket & return qwords written
   	iQWC=BuildGSPacket(pMemory->pvStart);

   	// Set the QWC register
   	*EE_D2_QWC=iQWC;

   	// Set the memory address register with the start of allocated memroy
   	*EE_D2_MADR=sps2GetPhysicalAddress(pMemory->pvStart, pMemory);

   	// Start the DMA transfer
   	*EE_D2_CHCR=chcr.i32;

   	// Wait for the DMA transfer to finish
   	sps2WaitForDMA(2, SPS2Descriptor);

}

bool BoxCollision(Ps2Box_t first, Ps2Box_t second)
{
	//It's easier to check when 2 boxes don't overlap:
	//They don't overlap if the right hand side of the the first box 
	//is left of the left hand side of the second
	if (first.x + first.w + first.xvel < second.x + second.xvel)
	{
		return false;
	}

	//They don't overlap if the left hand side of the first box is 
	//right of the right hand side of the second
	if (first.x + first.xvel > second.x + second.w + second.xvel)
	{
		return false;
	}

	//They don't overlap if the bottom of the first box is above the
	//top of the second (REMEMBER THE INVERTED y CO-ORDINATE!)
	if (first.y + first.yvel < second.y - second.h + second.yvel)
	{
		return false;
	}

	//They don't overlap if the top of the first box is below the 
	//bottom of the second
	if (first.y - first.h + first.yvel > second.y + second.yvel)
	{
		return false;
	}

	//If we've got this far, there is definately an overlap 
	//somewhere
	return true;	
}

void CollisionDetectionAndResponse(void)
{
	//Keep track of lives so we know when to end the game
	static int lives1 = 3;
	static int lives2 = 3;

	if (BoxCollision(Box[BALL], Box[LEFT_BAT]) || BoxCollision(Box[BALL],Box[RIGHT_BAT]))
	{
		//Collision occured between ball and a bat so turn ball around
                //Note - doesn't matter which bat.
		Box[BALL].xvel = -Box[BALL].xvel;
	}

	if (BoxCollision(Box[BALL],Box[UPPER_WALL]) || BoxCollision(Box[BALL],Box[LOWER_WALL]))
	{
		//Ball has hit upper or lower wall, so reverse y velocity
		//Note: It doesn't matter which wall is hit.
		Box[BALL].yvel = -Box[BALL].yvel;
	}

	if (!BoxCollision(Box[BALL],Box[BOX_INNER]))
	{
		//Ball has got past a bat - but which one?
		if (Box[BALL].x < 0)
		{
			//Left bat was beaten so Player 1 loses a life
			lives1--;
			switch (lives1)
			{
			case 2:
				// 1st life gone, blank out 3rd life box
				Box[LEFT_LIFE_3].r = 0;
				Box[LEFT_LIFE_3].g = 0;
				Box[LEFT_LIFE_3].b = 0;
			break;
			case 1:
				// 2nd life gone, blank out 2nd life box
				Box[LEFT_LIFE_2].r = 0;
				Box[LEFT_LIFE_2].g = 0;
				Box[LEFT_LIFE_2].b = 0;
			break;
			case 0:
				// 3rd life gone - just quit (Game over)
				exit(1);
			break;
			}
		}
		else
		{
			//Right Bat was beaten so player 2 loses a life
			lives2--;
			switch (lives2)
			{
			case 2:
				// 1st life gone, blank out 3rd life box
				Box[RIGHT_LIFE_3].r = 0;
				Box[RIGHT_LIFE_3].g = 0;
				Box[RIGHT_LIFE_3].b = 0;
			break;
			case 1:
				// 2nd life gone, blank out 2nd life box
				Box[RIGHT_LIFE_2].r = 0;
				Box[RIGHT_LIFE_2].g = 0;
				Box[RIGHT_LIFE_2].b = 0;
			break;
			case 0:
				// 3rd life gone - just quit (Game over)
				exit(1);
			break;
			}
		}

		//Regardless of which bat was beaten...
		//Reset counter to get a new 1 second pause
		counter = 0;
		//Reset ball position
		Box[BALL].x = -2;
		Box[BALL].y = 2;
		//I chose not to reset ball velocity, so ball will travel towards the player who just lost a life
	}
}

void DoControls(void)
{
	//Set speed of bat
	Box[LEFT_BAT].yvel = 2;
	//If X pressed, quadruple speed of bat
	if (pad[0].buttons & PAD_CROSS)
	{
		Box[LEFT_BAT].yvel = 8;
	}
	//If bat is between upper and lower walls, move bat in correct direction
	if (pad[0].buttons & PAD_UP)
	{
		Box[LEFT_BAT].yvel = -Box[LEFT_BAT].yvel;
		if (!BoxCollision(Box[LEFT_BAT],Box[UPPER_WALL]))
		{
			Box[LEFT_BAT].y += Box[LEFT_BAT].yvel;
		}
		else
		{
			//Ensure bat doesn't stop short
			Box[LEFT_BAT].y = Box[UPPER_WALL].y + Box[LEFT_BAT].h;
		}
	}
	if (pad[0].buttons & PAD_DOWN)
	{
		if (!BoxCollision(Box[LEFT_BAT],Box[LOWER_WALL]))
		{
			Box[LEFT_BAT].y += Box[LEFT_BAT].yvel;
		}
		else
		{
			//Ensure bat doesn't stop short
			Box[LEFT_BAT].y = Box[LOWER_WALL].y - Box[LOWER_WALL].h;
		}
	}
	
	//Same code for other bat
	Box[RIGHT_BAT].yvel = 2;
	if (pad[1].buttons & PAD_CROSS)
	{
		Box[RIGHT_BAT].yvel = 8;
	}
	if (pad[1].buttons & PAD_UP)
	{
		Box[RIGHT_BAT].yvel = -Box[RIGHT_BAT].yvel;
		if (!BoxCollision(Box[RIGHT_BAT],Box[UPPER_WALL]))
		{
			Box[RIGHT_BAT].y += Box[RIGHT_BAT].yvel;
		}
		else
		{
			Box[RIGHT_BAT].y = Box[UPPER_WALL].y + Box[RIGHT_BAT].h;
		}
	}
	if (pad[1].buttons & PAD_DOWN)
	{
		if (!BoxCollision(Box[RIGHT_BAT],Box[LOWER_WALL]))
		{
			Box[RIGHT_BAT].y += Box[RIGHT_BAT].yvel;
		}
		else
		{
			Box[RIGHT_BAT].y = Box[LOWER_WALL].y - Box[LOWER_WALL].h;
		}
	}
}

void UpdateBall(void)
{
	//counter is reset to zero after ball is lost, so if counter<50, we're in the first second of play (on a 50Hz display...).
	if (counter>50)
	{
		//At least a second has passed, so move the ball
		Box[BALL].x+=Box[BALL].xvel;
		Box[BALL].y+=Box[BALL].yvel;
	}
}

// Perform some game Logic
void GameLogic(void)
{
	CollisionDetectionAndResponse();
	DoControls();
	UpdateBall();
	counter++;
}
